import logging
from pathlib import Path
from typing import Optional

import typer
from typing_extensions import Annotated

from yamlex.api.splitter import split_yaml
from yamlex.api.util import (
    adjust_root_logger,
    get_default_extension_dir_path,
    get_default_extension_source_dir_path,
    is_manually_created,
    write_file,
)
from yamlex.cli.common_flags import (
    no_comment_flag,
    force_flag,
    verbose_flag,
    quiet_flag,
)


logger = logging.getLogger(__name__)


def split(
    source: Annotated[
        Optional[Path],
        typer.Option(
            "--source",
            "-s",
            help="Path to source extension.yaml file.",
            show_default="[default: extension/extension.yaml or src/extension/extension.yaml]",
            dir_okay=False,
            file_okay=True,
            exists=True,
            readable=True,
        ),
    ] = None,
    target: Annotated[
        Optional[Path],
        typer.Option(
            "--target",
            "-t",
            help="Path to directory where split YAML source files will be stored.",
            show_default="[default: source or src/source]",
            dir_okay=True,
            file_okay=False,
            exists=True,
            writable=True,
        ),
    ] = None,
    no_comment: no_comment_flag = False,
    force: force_flag = False,
    verbose: verbose_flag = False,
    quiet: quiet_flag = False,
) -> None:
    """
    Split [i]extension.yaml[/i] file into individual components.

    Performs a "best effort" opinionated splitting. This operation does not
    affect the original extension.yaml file. Instead, it extracts components
    from it and places them into individual files within the --target folder.

    [b]Splitting multiple times[/b]:

    Theoretically, you only split once. But in practice you can do it over
    and over (not very well tested). When performing consequent splitting,
    the operation overwrites any previously generated split files, if they
    have the 'Generated by yamlex' line within them. If the target file
    does not have the comment, it is considered to be manually created and is
    not overwritten. You can still force the overwrite using the --force flag.

    [b]Do not add 'Generated with yamlex' to files when splitting[/b]:

    When splitting, you can choose to not add the 'Generated by yamlex'
    comment to the generated files by using the --no-comment flag.
    """
    adjust_root_logger(verbose, quiet)

    source = source or get_default_extension_dir_path() / "extension.yaml"
    logger.debug(f"Source file: {source}")

    target = target or get_default_extension_source_dir_path()
    logger.debug(f"Target directory: {target}")

    split_parts = split_yaml(source, target)

    for path, part in split_parts.items():
        if is_manually_created(path) and not force:
            logger.info(f"(Skipping) Part: {path}")
        else:
            should_add_comment = not no_comment
            write_file(path, part, add_comment=should_add_comment)
            logger.info(f"Part written: {path}")



import logging
from typing import Optional
from typing_extensions import Annotated

import typer
from rich import print

from yamlex.cli.commands.join import join as join_cmd
from yamlex.cli.commands.map import map as map_cmd
from yamlex.cli.commands.split import split as split_cmd
from yamlex.cli.version import get_version


logging.basicConfig(level=logging.INFO, format="%(message)s")
logger = logging.getLogger(__name__)


def callback(
    ctx: typer.Context,
    version: Annotated[
        Optional[bool],
        typer.Option(
            "--version",
            "-v",
            show_default=False,
            is_eager=True,
        ),
    ] = None,
) -> None:
    if version:
        prog_version = get_version()
        print(prog_version)

        raise typer.Exit()
    
    if ctx.invoked_subcommand:
        return


def run() -> None:
    # Set up Typer app.
    app = typer.Typer(
        name="yamlex",
        rich_markup_mode="rich",
        add_completion=False,
        context_settings={"help_option_names": ["-h", "--help"]},
    )

    # Set primary callback. This shows helps or allows to show version.
    app.callback(invoke_without_command=True, no_args_is_help=True)(callback)

    # Declare main commands.
    app.command(name="map")(map_cmd)
    app.command(name="split")(split_cmd)
    app.command(name="join")(join_cmd)

    # Declare aliases for popular commands.
    app.command(name="j", hidden=True)(join_cmd)
    app.command(name="s", hidden=True)(split_cmd)

    # Launch Typer app.
    app()



from typing import Optional

import typer
from typing_extensions import Annotated


force_flag = Annotated[
    bool,
    typer.Option(
        "--force",
        "-f",
        help="Overwrite target files even if they were created manually.",
    ),
]
no_comment_flag = Annotated[
    bool,
    typer.Option(
        "--no-comment",
        "-C",
        help="Do not add the 'generated by yamlex' comment at the top of the file.",
    ),
]
verbose_flag = Annotated[
    bool,
    typer.Option(
        "--verbose",
        help="Enable verbose output.",
    ),
]
quiet_flag = Annotated[
    bool,
    typer.Option(
        "--quiet",
        help="Disable any informational output. Only errors.",
    ),
]
debug_flag = Annotated[
    bool,
    typer.Option(
        "--debug",
        hidden=True,
    ),
]



import logging
import re
from pathlib import Path
from typing import Union

import ruamel.yaml


logger = logging.getLogger(__name__)

DATASOURCE_NAMES = [
    "gcp",
    "jmx",
    "python",
    "processes",
    "snmp",
    "snmptraps",
    "sqlDb2",
    "sqlHana",
    "sqlMySql",
    "sqlOracle",
    "sqlPostgres",
    "sqlServer",
    "sqlSnowflake",
    "wmi",
]


def split_yaml(
    source_file_path: Path,
    target_dir_path: Path,
) -> dict[Path, Union[dict, list]]:
    """Decompose a YAML file into multiple files."""
    parts_to_write: dict[Path, Union[dict, list]] = dict()

    logger.info(f"Decomposing the central YAML file into parts: {source_file_path}")
    with open(source_file_path, "r") as extension_yaml_file:
        parser = ruamel.yaml.YAML()
        data: dict = parser.load(extension_yaml_file)

    # Process datasource
    logger.info(f"Processing datasource...")
    datasource = extract_datasource(data, target_dir_path)
    parts_to_write.update(datasource)

    # Process metrics
    logger.info(f"Processing metrics...")
    metrics_dir_path = target_dir_path / "metrics"
    metrics = extract_metrics(data, metrics_dir_path)
    parts_to_write.update(metrics)

    # Process screens
    logger.info(f"Processing screens...")
    screens_dir_path = target_dir_path / "screens"
    screens = extract_screens(data, screens_dir_path)
    parts_to_write.update(screens)

    # Process topology
    logger.info(f"Processing topology...")
    topology_dir_path = target_dir_path / "topology"
    topology = extract_topology(data, topology_dir_path)
    parts_to_write.update(topology)
    
    # Dump the rest into the index file
    index_yaml_file_path = target_dir_path / "index.yaml"
    parts_to_write[index_yaml_file_path] = data
    logger.info(f"Index file extracted: {index_yaml_file_path}")

    # Return collected summary
    return parts_to_write


def extract_datasource(
    extension: dict,
    target_dir_path: Path
) -> dict[Path, dict]:
    paths_to_write = {}

    # Detect the datasource name
    datasource_name: str | None = None
    datasource: dict | list | None = None
    for name in DATASOURCE_NAMES:
        if name in extension:
            datasource_name = name
            datasource = extension.pop(datasource_name)
            break

    if datasource_name:
        logger.info(f"Datasource name detected: {datasource_name}")
    else:
        logger.error("Datasource definition not found")
        return paths_to_write
    
    datasource_dir_path = target_dir_path / datasource_name
    
    if datasource_name == "python":
        python_index_yaml_path = datasource_dir_path / "index.yaml"
        paths_to_write[python_index_yaml_path] = datasource

    elif datasource_name == "processes":
        for process in datasource:
            process_name = process.get("name")
            if not process_name:
                logger.error("Process does not have a name")
                continue

            process_file_name = sanitize_file_stem(process_name)
            process_file_path = datasource_dir_path / f"-{process_file_name}.yaml"
            paths_to_write[process_file_path] = process
            logger.info(f"Process extracted: {process_file_path}")

    else:
        for group in datasource:
            group_name = group.get("group")
            if not group_name:
                logger.error("Group does not have a name")
                continue

            subgroups = group.pop("subgroups", None)
            # If subgroups exist, split them into different files
            if isinstance(subgroups, list):
                group_file_name = sanitize_file_stem(group_name)
                group_dir_path = datasource_dir_path / "groups" / group_file_name
                for subgroup in subgroups:
                    subgroup_name = subgroup.get("subgroup")
                    if not subgroup_name:
                        logger.error("Subgroup does not have a name")
                        continue

                    subgroup_file_name = sanitize_file_stem(subgroup_name)
                    subgroup_file_path = group_dir_path / "subgroups" / f"-{subgroup_file_name}.yaml"
                    paths_to_write[subgroup_file_path] = subgroup
                    logger.info(f"Subgroup extracted: {subgroup_file_path}")

                group_file_path = group_dir_path / "index.yaml"
                paths_to_write[group_file_path] = group
                logger.info(f"Group extracted: {group_file_path}")

            # If subgroups do not exist, write each group into
            # individual files.
            else:
                group_file_name = sanitize_file_stem(group_name)
                group_file_path = datasource_dir_path / f"-{group_file_name}.yaml"
                paths_to_write[group_file_path] = group
                logger.info(f"Group extracted: {group_file_path}")

    return paths_to_write


def extract_metrics(extension: dict, metrics_dir_path: Path) -> dict[Path, dict]:
    paths_to_write = {}
    if not isinstance(extension.get("metrics"), list):
        return paths_to_write

    metrics: list[dict] = extension.pop("metrics")
    for metric in metrics:
        metric_key = metric.get("key")
        if not metric_key:
            logger.error(f"Metric does not have a 'key': {metric}")
            continue

        metric_file_name = sanitize_file_stem(metric_key)

        # Extract metric into a separate file
        metric_file_path = metrics_dir_path / f"-{metric_file_name}.yaml"
        paths_to_write[metric_file_path] = metric
        logger.info(f"Metric extracted: {metric_file_path}")

    return paths_to_write


def extract_screens(extension: dict, screens_dir_path: Path) -> dict[Path, dict]:
    paths_to_write = {}
    if not isinstance(extension.get("screens"), list):
        return paths_to_write

    screens: list[dict] = extension.pop("screens")
    for screen in screens:
        entity_type: str = screen.get("entityType")
        if not entity_type:
            logger.error(f"Screen does not have an entity_type")
            continue
            
        screen_file_name = sanitize_file_stem(entity_type)

        # Extract screen into a separate file
        screen_file_path = screens_dir_path / f"-{screen_file_name}.yaml"
        paths_to_write[screen_file_path] = screen
        logger.info(f"Screen extracted: {screen_file_path}")

    return paths_to_write


def extract_topology(extension: dict, topology_dir_path: Path) -> dict[Path, dict]:
    paths_to_write = {}
    if not isinstance(extension.get("topology"), dict):
        return paths_to_write
    
    topology: dict = extension.pop("topology")

    types_dir_path = topology_dir_path / "types"
    types = extract_types(topology, types_dir_path)
    paths_to_write.update(types)

    relationships_dir_path = topology_dir_path / "relationships"
    relationships = extract_relationships(topology, relationships_dir_path)
    paths_to_write.update(relationships)

    return paths_to_write


def extract_types(topology: dict, types_dir_path: Path) -> dict[Path, dict]:
    paths_to_write = {}
    if not isinstance(topology.get("types"), list):
        return paths_to_write
    
    types: list[dict] = topology.pop("types")
    for type_ in types:
        type_name = type_.get("name")
        if not type_name:
            logger.error(f"Type does not have a name: {type_}")
            continue

        type_file_name = sanitize_file_stem(type_name)

        # Extract type into a separate file
        type_file_path = types_dir_path / f"-{type_file_name}.yaml"
        paths_to_write[type_file_path] = type_
        logger.info(f"Type extracted: {type_file_path}")

    return paths_to_write


def extract_relationships(topology, relationships_dir_path: Path) -> dict[Path, dict]:
    paths_to_write = {}
    if not isinstance(topology.get("relationships"), list):
        return paths_to_write
    
    relationships: list[dict] = topology.pop("relationships")
    for relationship in relationships:
        if (
            "fromType" not in relationship
            or "toType" not in relationship
            or "typeOfRelation" not in relationship
        ):
            logger.error(f"Relationship does not have required fields: {relationship}")
            continue

        relationship_file_name = sanitize_file_stem(
            (
                f"{relationship['fromType']}_"
                f"{relationship['typeOfRelation']}_"
                f"{relationship['toType']}"
            ),
        )

        # Extract relationship into a separate file
        relationship_file_path = relationships_dir_path / f"-{relationship_file_name}.yaml"
        paths_to_write[relationship_file_path] = relationship
        logger.info(f"Relationship extracted: {relationship_file_path}")

    return paths_to_write


def sanitize_file_stem(stem: str) -> Path:
    valid_id = re.sub(r"[^a-z0-9\.]", "_", stem.lower())
    return valid_id



import logging
import re
from io import StringIO
from pathlib import Path
from typing import Optional, Union

import ruamel.yaml


logger = logging.getLogger(__name__)
parser = ruamel.yaml.YAML()


def adjust_root_logger(verbose: bool = False, quiet: bool = False) -> None:
    if quiet:
        logging.root.setLevel(logging.ERROR)
    elif verbose:
        logging.root.setLevel(logging.DEBUG)


def get_default_extension_source_dir_path() -> Path:
    src_dir_path = Path("src")
    if src_dir_path.exists() and src_dir_path.is_dir():
        return src_dir_path / "source"
    else:
        return Path("source")


def get_default_extension_dir_path() -> Path:
    candidates = [
        Path("extension"),
        Path("src/extension"),
    ]

    # Find the first suitable candidate
    for candidate in candidates:
        # Check the candidate directory exists
        if candidate.exists() and candidate.is_dir():
            return candidate
        continue

    # Nothing suitable was found
    logger.error(
        (
            "No default extension directory found. "
            "'extension/' or 'src/extension/' folder must exist "
            "in the current working directory."
        )
    )
    exit(1)


def is_manually_created(path: Path) -> bool:
    if path.exists():
        with open(path, "r") as f:
            content = f.read().lower()
            if "generated by yamlex" not in content:
                True
    return False


def write_file(
    file_path: Path,
    data: Union[dict, list],
    add_comment: bool = True,
) -> None:
    # Convert dict to YAML. Dump to string first to add a comment
    stream = StringIO()
    parser.indent(mapping=2, sequence=4, offset=2)
    parser.dump(data, stream)
    text = stream.getvalue()

    # Make sure the directory we write to exists
    dir = file_path.parent.resolve()
    dir.mkdir(parents=True, exist_ok=True)

    # Write to output file
    with open(file_path, "w") as f:
        if add_comment:
            # Write a comment to indicate that the file was automatically generated
            comment = f"# Generated by yamlex\n\n"
            f.write(comment)
        f.write(text)


def read_version_properties(path: Path, default: Optional[str] = None) -> str:
    try:
        with open(path, "r") as f:
            content = f.read()
            match_ = re.search(r"version=(.*)", content)
            if not match_ and not default:
                raise KeyError("Could not read current version")
            return match_.group(1)
    except Exception as e:
        logger.error(f"Failed to open {path}: {e}")
        exit(1)


def write_version_properties(path: Path, version: str) -> None:
    try:
        with open(path, "r") as rf:
            lines = rf.readlines()
            content = []
            for l in lines:
                if l.startswith("version"):
                    content.append(f"version={version}\n")
                else:
                    content.append(l)
        with open(path, "w") as wf:
            wf.writelines(content) 
    except Exception as e:
        logger.error(f"Failed to write to {path}: {e}")
        exit(1)


def parse_version(version: str) -> Optional[str]:
    regex = r"^(0|[1-9][0-9]{0,3})(\.(0|[1-9][0-9]{0,3})){0,2}$"
    match_ = re.search(regex, version)
    if not match_:
        return None
    return match_.group()


def bump_version(version: str) -> str:
    version_parts = version.split(".")
    part_int: int | None = None
    part_idx: int | None = None
    for i, part in enumerate(reversed(version_parts)):
        try:
            part_int = int(part)
            part_idx = len(version_parts) - 1 - i
            break
        except:
            continue
    if part_int is None:
        logger.error("Could not find integer part in version")
        exit(1)

    part_int += 1
    result = ""
    for i, part in enumerate(version_parts):
        append = part
        if i == part_idx:
            append = part_int
        result += f"{append}."
    result = result.rstrip(".")
    return result



import logging
from pathlib import Path
from typing import Optional

import typer
from typing_extensions import Annotated

from yamlex.api.joiner import (
    assemble_recursively,
    remove_yaml_comments,
)
from yamlex.api.util import (
    adjust_root_logger,
    get_default_extension_dir_path,
    get_default_extension_source_dir_path,
    is_manually_created,
    write_file,
    read_version_properties,
    parse_version,
)
from yamlex.cli.common_flags import (
    no_comment_flag,
    force_flag,
    verbose_flag,
    quiet_flag,
    debug_flag,
)


logger = logging.getLogger(__name__)


def join(
    source: Annotated[
        Optional[Path],
        typer.Option(
            "--source",
            "-s",
            help="Path to the directory where individual source component files are stored.",
            show_default="[default: source or src/source]",
            dir_okay=True,
            file_okay=False,
            exists=True,
            readable=True,
        )
    ] = None,
    target: Annotated[
        Optional[Path],
        typer.Option(
            "--target",
            "-t",
            help="Path to the target extension.yaml file that will be assembled from parts.",
            show_default="[default: extension/extension.yaml or src/extension/extension.yaml]",
            dir_okay=False,
            file_okay=True,
        )
    ] = None,
    dev: Annotated[
        bool,
        typer.Option(
            "--dev",
            "-d",
            help="Add 'custom:' prefix and explicit version when producing extension.yaml.",
        ),
    ] = False,
    keep_formating: Annotated[
        bool,
        typer.Option(
            "--keep",
            "-k",
            help="Keep formatting and indentation when adding non-yaml files into extension.yaml.",
        ),
    ] = True,
    version: Annotated[
        Optional[str],
        typer.Option(
            "--version",
            "-v",
            help="Explicitly set the version in the extension.yaml.",
            show_default=False,
        ),
    ] = None,
    no_comment: no_comment_flag = False,
    force: force_flag = False,
    verbose: verbose_flag = False,
    quiet: quiet_flag = False,
    debug: debug_flag = False,
) -> None:
    """
    Join individual components into a single [i]extension.yaml[/i] file.

    Assembles all files from the --source directory in a hierarchical order.
    As if the folder structure of the source directory represents a YAML
    structure.

    [b]How it works[/b]

    - Any folder or file within the --source directory is considered to be
      a field within the final extension.yaml. For example, a file called
      [i]name.yaml[/i] will become a field called [i]name:[/i] and its content will be
      put into that field.

    - Nesting level of the included file or folder matches its
      indentation level within the resulting file.

    - If a folder contains a special file called [i]index.yaml[/i], then the
      content of that file is added on the same level as the folder. For
      example, 

    - If any of the items within the folder start with the minus sign "-" in
      their name, then the whole folder is considered to be an array.

    - If a folder name starts with a plus sign "+", then the folder is
      considered to be a grouper folder and yamlex behaves as if that folder
      does not add any additional level of nesting at all.

    [b]Full example:[/b]

    [i]Source folder structure:[/i]          [i]Content of files on the left:[/i]

    source:
    ├── metadata.yaml                 name: My name is yamlex
    ├── folder
    │   ├── query.sql                 SELECT * FROM DUAL
    │   ├── !some.yaml                text: This file will be skipped
    │   └── index.yaml                group: My Query
    ├── +grouper
    │   └── normal.yaml               present: value
    └── array
        ├── -item_1.yaml              call: fus
        ├── -item_2.yaml              call: roh
        └── -item_3
             └── index.yaml           call: dah

    [i]Resulting extension.yaml:[/i]

    metadata:
      name: My name is yamlex
    folder:
      group: My Query
      query: SELECT * FROM DUAL
    present: value
    array:
      - call: fus
      - call: roh
      - call: dah

    [b]Overwriting existing extension.yaml (--no-comment and --force)[/b]

    Yamlex tries to be cautious not to accidentally overwrite a manually
    created [i]extension.yaml[/i]. If that file contains the "Generated with
    yamlex" line in it, then yamlex overwrites it without hesitation.
    However, when [i]extension.yaml[/i] does not contain that line, yamlex
    does not overwrite it. You can alter this behaviour using --force flag.
    
    When yamlex generates the [i]extension.yaml[/i] from parts, it adds the
    same comment at the top: # Generated by yamlex
    If you would like to disable this behaviour, use the --no-comment flag.

    [b]Development mode[/b]

    When you add the --dev flag, yamlex will add the "custom:" prefix to the
    name of your extension and will put an explicit version into the final
    [i]extension.yaml[/i].
    """
    adjust_root_logger(verbose, quiet)

    source = source or get_default_extension_source_dir_path()
    logger.debug(f"Source files directory: {source}")

    try:
        extension = assemble_recursively(
            source,
            non_yaml_files_as_scalars=keep_formating,
            debug=debug,
        )
    except UnicodeDecodeError as e:
        # If any non-text files were caught up in the processing because they
        # happened to be placed into the source directory, we can catch it
        # here, since these files would file to be decoded as utf-8.
        logger.error(f"Found non-text file in the source directory: {e}")
        raise typer.Exit(1)

    # Figure out the current version
    yaml_version = extension.get("version")

    if dev:
        name = extension.get("name")
        if isinstance(name, str) and not name.startswith("custom:"):
            name = f"custom:{name}"
            extension["name"] = name
            logger.info(f"Dev mode extension name: {name}")

        # If new version is not explicitly specified
        if not version:
            current_version = parse_version(yaml_version)
            if not current_version:
                current_version = read_version_properties("version.properties")
                logger.info(f"Version found in version.properties: {current_version}")

        # Update version, if new explicit version was specified
        if not version:
            version = current_version

        # Explicitly add version to generated extension.yaml
        if yaml_version != version:
            extension["version"] = version
            logger.info(f"Explicit version in dev mode: {version}")

    target = target or get_default_extension_dir_path() / "extension.yaml"
    logger.debug(f"Target file: {target}")

    # Check if the target file exists and was created manually
    if is_manually_created(target) and not force:
        logger.error(f"The {target} file was created manually. Use --force to overwrite it.")
        exit(2)

    # Remove comments
    remove_yaml_comments(extension)

    # Write to output file
    should_add_comment = not no_comment
    write_file(target, extension, add_comment=should_add_comment)

    if debug:
        print(extension)
